import  java.util.Arrays;
import  java.util.HashMap;
import  java.util.Map;

/**
 * Dynamically growing bit set.
 * @author rschuetz
 */

public class DynBitSet
{
 // === Static ================================================================

 private static final int  ELEMENT_BITS_LOG2 = 5,                      // log2(32), 32bits / integer
                           ELEMENT_BITS      = 1 << ELEMENT_BITS_LOG2; // 2^5, 32bits / integer

 private static final Page PAGE_ALL_BITS_SET = new DummyPage();

 // === Attributes ============================================================

 private final Map<Long, Page> pages;
 private final int             pageSize;
 private       long            nbrBitsSet;
 
 // === Constructor ===========================================================

 /**
  * Create a new DynBitSet instance
  * @param pageSize the page size in bits
  */

 public DynBitSet(final int pageSize)
 {
  // Asserts

  assert pageSize % ELEMENT_BITS == 0L;

  // Initialize

  this.pages      = new HashMap<Long, Page>();
  this.pageSize   = pageSize;
  this.nbrBitsSet = 0L;
 }

 // === Public ================================================================

 /**
  * @param position the position
  * @return the value of a bit at the specified position
  */

 public boolean get(final long position)
 {
  Page page;

  // Get page

  page = pages.get(position / pageSize);

  // Perform special handling based on the returned page.
  // The contract is the following:
  //
  // page == null              : all bits cleared
  // page == PAGE_ALL_BITS_SET : all bits set

  return(page == null              ? false :
        (page == PAGE_ALL_BITS_SET ? true  : page.get(((int)(position % pageSize)))));
 }

 /**
  * Set bit at the specified position
  * @param position the position
  * @return true, if the bit has been set;
  *         false, if it was already set
  */

 public boolean set(final long position)
 {
  boolean bitSet;
  long    pageIndex;

  Page    page;

  // Calculate page index

  pageIndex = position / pageSize;

  // Get page; create it with no bits set if not present.

  if((page = pages.get(pageIndex)) == null)
  {
   pages.put(pageIndex, page = new StandardPage(pageSize, false));
  }

  // Proceed based on returned page

  if(page == PAGE_ALL_BITS_SET)
  {
   // Page is a dummy page indicating that all its
   // bits are already set

   bitSet = false;
  }
  else
  {
   // Standard page; try to set the bit

   if((bitSet = page.set(((int)(position % pageSize)))))
   {
    // Bit has just been set; adapt counter

    nbrBitsSet ++;

    // Replace page (to save space), if all bits are set.
    
    if(page.hasAllBitsSet()) pages.put(pageIndex, PAGE_ALL_BITS_SET);
   }
  }

  // Return result

  return(bitSet);
 }

 /**
  * Clear bit at the specified position
  * @param position the position
  * @return true, if the bit has been cleared;
  *         false, if it wasn't set
  */

 public boolean clear(final long position)
 {
  long    pageIndex;
  boolean bitCleared;

  Page    page;

  // Calculate page index

  pageIndex = position / pageSize;

  // Get page

  page = pages.get(pageIndex);

  // Proceed based on returned page

  if(page == null)
  {
   // Page does not exists, no bits are set

   bitCleared = false;
  }
  else if(page == PAGE_ALL_BITS_SET)
  {
   // Page is a dummy page indicating that all its
   // bits are already set. Replace page with real
   // page with all bits set

   pages.put(pageIndex, page = new StandardPage(pageSize, true));

   // Clear the specified bit, adapt counter. 

   if((bitCleared = page.clear(((int)(position % pageSize)))))
   {
    // Bit has just been cleared; adapt counter

    nbrBitsSet --;
   }
  }
  else
  {
   // Standard page; try to clear the bit
   
   if((bitCleared = page.clear(((int)(position % pageSize)))))
   {
    // Bit has just been cleared; adapt counter

    nbrBitsSet --;
    
    // Remove page, to save space, if all bits are cleared.

    if(page.hasAllBitsCleared()) pages.remove(pageIndex);
   }
  }

  // Clear result

  return(bitCleared);
 }

 /**
  * @return the number of set bits
  */

 public long getNbrSetBits()
 {
  return(nbrBitsSet);
 }

 // === Package protected =====================================================
 // === Protected =============================================================
 // === Private ===============================================================

 // === Inner classes =========================================================

 private interface Page
 {
  /**
   * @param position the position
   * @return the value of a bit at the specified position
   */

  public boolean get(final int position);

  /**
   * Set bit at the specified position
   * @param position the position
   * @return true, if the bit has been set; false, if it was already set
   */

  public boolean set(final int position);

  /**
   * Clear bit at the specified position
   * @param position the position
   * @return true, if the bit has been cleared;
   *         false, if it wasn't set
   */

  public boolean clear(final int position);

  /**
   * @return the number of set bits
   */

  public int getNbrBitsSet();
  
  /**
   * @return the number of set bits
   */

  public int getNbrBitsCleared();  
  
  /**
   * @return the true, if the page has all bits set
   */

  public boolean hasAllBitsSet();

  /**
   * @return true, if the page has all bits cleared
   */

  public boolean hasAllBitsCleared();

  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================
 }

 /**
  * Standard page implementation
  */

 private static class StandardPage implements Page
 {
  // === Static ================================================================
  // === Attributes ============================================================

  private final int bits[];
  private       int size, nbrBitsSet;

  // === Constructor ===========================================================

  /**
   * Create a new StandardPage instance
   * @param size    the number of bits this page holds; must be a multiple of ELEMENT_BITS
   * @param setBits if true, set all bits of the page; if false clear them
   */

  public StandardPage(final int size, boolean setBits)
  {
   // Asserts

   assert (size & (ELEMENT_BITS - 1)) == 0;
   assert  size                       >  0;

   // Initialize

   this.bits       = new int[size >>> ELEMENT_BITS_LOG2];
   this.size       = size;
   this.nbrBitsSet = 0;

   // Initialize bits

   Arrays.fill(bits, setBits ? 0xFFFFFFFF : 0x00000000);
  }

  // === Public ================================================================

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#get(int)
   */

  @Override
  public boolean get(final int position)
  {
   int index, bitMask;

   // Asserts

   assert position >= 0 && position < bits.length * ELEMENT_BITS;

   // Calculate index and bit mask

   index   =       position >>> ELEMENT_BITS_LOG2;
   bitMask = 1 << (position &  (ELEMENT_BITS - 1));

   // Get and return

   return((bits[index] & bitMask) != 0);
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#set(int)
   */

  @Override
  public boolean set(final int position)
  {
   int     index, bitMask;
   boolean changed;

   // Asserts

   assert position >= 0 && position < bits.length * ELEMENT_BITS;

   // Calculate index and bit mask

   index   =       position >>> ELEMENT_BITS_LOG2;
   bitMask = 1 << (position &  (ELEMENT_BITS - 1));

   // Get & Set

   if(changed = ((bits[index] & bitMask) == 0))
   {
    // Set bit

    bits[index] |= bitMask;

    // Update counter

    nbrBitsSet ++;
   }

   // Return

   return(changed);
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#clear(int)
   */

  public boolean clear(final int position)
  {
   int     index, bitMask;
   boolean changed;

   // Asserts

   assert position >= 0 && position < bits.length * ELEMENT_BITS;

   // Calculate index and bit mask

   index   =       position >>> ELEMENT_BITS_LOG2;
   bitMask = 1 << (position &  (ELEMENT_BITS - 1));

   // Get & Clear

   if(changed = ((bits[index] & bitMask) != 0))
   {
    // Clear bit

    bits[index] &= ~bitMask;

    // Update counter

    nbrBitsSet --;
   }

   // Return

   return(changed);
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#getNbrBitsSet()
   */
  
  @Override  
  public int getNbrBitsSet()
  {
   return(nbrBitsSet);
  }
  
  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#getNbrBitsCleared()
   */
  
  @Override
  public int getNbrBitsCleared()
  {
   return(size - nbrBitsSet);
  }
  
  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#hasAllBitsSet()
   */
  
  @Override
  public boolean hasAllBitsSet()
  {
   return(nbrBitsSet == size);
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#hasAllBitsCleared()
   */
  
  @Override
  public boolean hasAllBitsCleared()
  {
   return(nbrBitsSet == 0);
  }

  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================
 }

 /**
  * Dummy page implementation
  */

 private static class DummyPage implements Page
 {
  // === Static ================================================================
  // === Attributes ============================================================
  // === Constructor ===========================================================

  /**
   * Create a new DummyPage instance
   */

  public DummyPage()
  {
  }

  // === Public ================================================================

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#get(int)
   */

  @Override
  public boolean get(final int position)
  {
   throw new UnsupportedOperationException();
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#set(int)
   */

  @Override
  public boolean set(final int position)
  {
   throw new UnsupportedOperationException();
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#clear(int)
   */

  public boolean clear(final int position)
  {
   throw new UnsupportedOperationException();
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#getNbrBitsSet()
   */
  
  @Override  
  public int getNbrBitsSet()
  {
   throw new UnsupportedOperationException();
  }
  
  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#getNbrBitsCleared()
   */
  
  @Override
  public int getNbrBitsCleared()
  {
   throw new UnsupportedOperationException();
  }
  
  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#hasAllBitsSet()
   */
  
  @Override
  public boolean hasAllBitsSet()
  {
   throw new UnsupportedOperationException();
  }

  /**
   * @see com.tomtomwork.webfleet.components.base.util.DynBitSet.Page#hasAllBitsCleared()
   */
  
  @Override
  public boolean hasAllBitsCleared()
  {
   throw new UnsupportedOperationException();
  }

  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================
 }
}
