import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import java.text.DecimalFormat;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.DeliveryMode;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.MapMessage;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;

import com.tomtomwork.webfleet.components.base.util.DynBitSet;

/**
 * Multithreaded load test
 * @author rschuetz
 */

public class MultiThreadedLoadTest
{
 // === Static ================================================================

 private static enum BatchMode { COUNT, TIME };

 private static final Logger         LOGGER          = Logger.getLogger(LoadTest.class);

 private static final long           RECEIVE_TIMEOUT  = 5L * 60L * 1000,
                                     MSG_LOG_INTERVAL = 5L * 1000L;

 private static final String         ID_MESSAGEID    = "messageid",
                                     ID_SEQUENCEID   = "sequenceid",
                                     ID_RETRYID      = "retryid",
                                     ID_TIMESTAMP    = "timestamp",
                                     ID_PAYLOAD      = "payload";

 private static final String         NBR_FORMAT      = "###,###,###,###,###,###";

 // === Attributes ============================================================

 private final Map<Integer, byte[]>  MESSAGES;

 private final Random                RANDOM;

 private final File                  STOPFILE;

 private final long                  STARTTIME;

 private final int                   MINMSGCOUNT,
                                     MAXMSGCOUNT,
                                     MINMSGSIZE,
                                     MAXMSGSIZE,
                                     SENDERCOMMITINTERVAL,
                                     RECEIVERCOMMITINTERVAL;

 private final BatchMode             SENDERBATCHMODE,
                                     RECEIVERBATCHMODE;

 private final boolean               PERSISTANT_STORAGE;

 private       Sender                sender;
 private       List<Receiver>        receivers;

 // === Constructor ===========================================================

 /**
  * Create a new MultiThreadedLoadTest instance
  * @param args the command line arguments
  */

 public MultiThreadedLoadTest(String[] args)
 {
  int                   argIdx;
  byte[]                message;

  String                senderPropertiesFile, receiverPropertiesFiles[];

  Map <Integer, byte[]> messages;

  // Initialize logger

  BasicConfigurator.resetConfiguration();

  Logger.getRootLogger().addAppender(new ConsoleAppender(new PatternLayout("%-5p [%d{ISO8601}] %m%n")));
  Logger.getRootLogger().setLevel(Level.INFO);

  // Get parameters

  if(args.length == 12)
  {
   argIdx                     = 0;

   senderPropertiesFile       =                      args[argIdx ++];
   receiverPropertiesFiles    =                      args[argIdx ++].split(",");

   STOPFILE                   = new File            (args[argIdx ++]);
   MINMSGCOUNT                = Integer.parseInt    (args[argIdx ++]);
   MAXMSGCOUNT                = Integer.parseInt    (args[argIdx ++]);
   MINMSGSIZE                 = Integer.parseInt    (args[argIdx ++]);
   MAXMSGSIZE                 = Integer.parseInt    (args[argIdx ++]);

   SENDERBATCHMODE            = BatchMode.valueOf   (args[argIdx ++]);
   SENDERCOMMITINTERVAL       = Integer.parseInt    (args[argIdx ++]);

   RECEIVERBATCHMODE          = BatchMode.valueOf   (args[argIdx ++]);
   RECEIVERCOMMITINTERVAL     = Integer.parseInt    (args[argIdx ++]);

   PERSISTANT_STORAGE         = Boolean.parseBoolean(args[argIdx ++]);
  }
  else
  {
   throw new IllegalArgumentException("Usage: " + MultiThreadedLoadTest.class.getSimpleName() +
                                      " <sender properties file> <receiver properties file(s); comma separated>" +
                                      " <stop file name> <min. message count in a queue> <max. message count in a queue>" +
                                      " <minimum message size in bytes> <maximum message size in bytes>" +
                                      " <sender batch mode (COUNT|TIME)> <sender commit interval in # of messages processed resp. milliseconds> " +
                                      " <receiver batch mode (COUNT|TIME)> <receiver commit interval in # of messages processed resp. milliseconds> " +
                                      " <persistant storage (true/false)>");
  }

  // Log parameters

  LOGGER.info("=== " + this.getClass().getSimpleName() + " ===");
  LOGGER.info("SENDERCONFIG               = " + senderPropertiesFile);
  LOGGER.info("RECEIVERCONFIGS            = " + Arrays.asList(receiverPropertiesFiles));
  LOGGER.info("STOPFILE                   = " + STOPFILE);
  LOGGER.info("MINMSGCOUNT                = " + MINMSGCOUNT);
  LOGGER.info("MAXMSGCOUNT                = " + MAXMSGCOUNT);
  LOGGER.info("MINMSGSIZE                 = " + MINMSGSIZE);
  LOGGER.info("MAXMSGSIZE                 = " + MAXMSGSIZE);
  LOGGER.info("SENDERBATCHMODE            = " + SENDERBATCHMODE);
  LOGGER.info("SENDERCOMMITINTERVAL       = " + SENDERCOMMITINTERVAL);
  LOGGER.info("RECEIVERBATCHMODE          = " + RECEIVERBATCHMODE);
  LOGGER.info("RECEIVERCOMMITINTERVAL     = " + RECEIVERCOMMITINTERVAL);
  LOGGER.info("PERSISTANT_STORAGE         = " + PERSISTANT_STORAGE);

  // Initialize

  RANDOM    = new Random();
  STARTTIME = System.currentTimeMillis();

  // Pre-create messages

  messages = new HashMap<Integer, byte[]>();

  for(int i = MINMSGSIZE; i <= MAXMSGSIZE; i ++)
  {
   message = new byte[i];
   RANDOM.nextBytes(message);
   messages.put(i, message);
  }

  MESSAGES = Collections.unmodifiableMap(messages);

  // Create sender

  sender = new Sender("S-", new File(senderPropertiesFile));

  // Create receivers

  receivers = new LinkedList<Receiver>();

  for(int receiverID = 0; receiverID < receiverPropertiesFiles.length; receiverID ++)
  {
   receivers.add(new Receiver("R" + receiverID, new File(receiverPropertiesFiles[receiverID])));
  }
 }

 // === Public ================================================================

 /**
  * @param args the command line arguments
  * @throws Exception
  */

 public static void main(String[] args)
 throws Exception
 {
  (new MultiThreadedLoadTest(args)).runLoadTest();
  System.exit(0);
 }

 /**
  * Run the load test
  */

 public void runLoadTest()
 {
  try
  {
   // Launch threads

                                      sender  .start();
   for(Receiver receiver : receivers) receiver.start();

   // Loop

   while(true)
   {
    // Check for stop file

    if(STOPFILE.exists())
    {
     LOGGER.info("Stop file found; stopping ...");
     break;
    }
    else
    {
     Thread.sleep(100);
    }
   }
   
   // Signal threads to stop

                                      sender  .signalStop();
   for(Receiver receiver : receivers) receiver.signalStop();

   // Join

                                      sender  .join();
   for(Receiver receiver : receivers) receiver.join();
  }
  catch(Exception e)
  {
   LOGGER.error("Exception caught", e);
  }
  finally
  {
   try
   {
                                       sender  .disconnect();
    for(Receiver receiver : receivers) receiver.disconnect();
   }
   catch(Exception e)
   {
    LOGGER.error("Exception caught while disconnecting", e);
   }
  }
 }
  
 private synchronized void controlThreads()
 {
  long backlog, maxBacklog;

  // Control sender and receivers

  maxBacklog = 0L;

  for(Receiver receiver : receivers)
  {
   // Get message count difference

   backlog = sender.getCurrentMessageSequenceID() - receiver.getCurrentMessageSequenceID();

   // Pause/Resume receiver

   receiver.signalPause(backlog < MINMSGCOUNT);

   // Get max. difference

   maxBacklog = Math.max(maxBacklog, backlog);
  }

  // Pause/Resume sender

  sender.signalPause(maxBacklog > MAXMSGCOUNT);
 }

 // === Package protected =====================================================
 // === Protected =============================================================
 // === Private ===============================================================

 /**
  * Load a properties file
  * @param file the file reference
  * @return the properties
  */

 Properties loadProperties(File file)
 {
  Properties  properties;
  InputStream inputStream;

  // Initialize JNDI

  properties  = new Properties();
  inputStream = null;

  try
  {
   inputStream = new FileInputStream(file);
   properties.load(inputStream);
  }
  catch(IOException e)
  {
   LOGGER.warn("Unable to read file \"" + file.getAbsolutePath() + "\".", e);
  }
  finally
  {
   if(inputStream != null) try { inputStream.close(); } catch(Exception e) {}
  }

  return(properties);
 }

 // === Inner classes =========================================================

 private abstract class Connector extends Thread
 {
  // === Static ================================================================
  // === Attributes ============================================================

           Connection          connection;
           Session             session;

  final    String              name;
  final    String              jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias;
  final    Map<String, String> addJndiICProperties;

  final    DecimalFormat       numberFormatter;

  volatile long                msgSequenceID;
  volatile boolean             stop, pause;

  // === Constructor ===========================================================

  /**
   * Create a new connector instance
   * @param name                       the connector name
   * @param jndiFactoryInitial         the JNDI factory alias
   * @param jndiProviderURL            the JNDI provider URL
   * @param jndiConnectionFactoryAlias the JNDI connection factory alias
   * @param jndiQueueAlias             the JNDI queue alias
   * @param addJndiICProperties        any additional JNDI InitialContext properties
   */

  public Connector(final String name,
                   final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                   final Map<String, String> addJndiICProperties)
  {
   // Initialize

   this.name             = name;
   this.msgSequenceID    = 0;
   this.pause            = false;
   this.stop             = false;
   this.numberFormatter  = new DecimalFormat(NBR_FORMAT);

   // Set properties

   this.jndiFactoryInitial         = jndiFactoryInitial;
   this.jndiProviderURL            = jndiProviderURL;
   this.jndiConnectionFactoryAlias = jndiConnectionFactoryAlias;
   this.jndiQueueAlias             = jndiQueueAlias;
   this.addJndiICProperties        = Collections.unmodifiableMap(addJndiICProperties);

   // Set thread name

   setName(getClass().getName() + " : " + name);

   // Log parameters

   LOGGER.info("=== " + this.getClass().getSimpleName() + " :: " + name + " ===");
   LOGGER.info("JNDIFACTORYINITIAL         = " + jndiFactoryInitial);
   LOGGER.info("JNDIPROVIDERURL            = " + jndiProviderURL);
   LOGGER.info("JNDICONNECTIONFACTORYALIAS = " + jndiConnectionFactoryAlias);
   LOGGER.info("JNDIQUEUEALIAS             = " + jndiQueueAlias);
   LOGGER.info("ADDJNDIICPROPERTIES        = " + addJndiICProperties);
  }

  /**
   * Create a new connector instance
   * @param name                 the connector name
   * @param connectionProperties a file reference pointing to a properties file containing
   *                             the connector properties
   */

  public Connector(final String name, final File connectionProperties)
  {
   Properties          properties;
   Map<String, String> addJndiICProperties;

   // Initialize

   this.name             = name;
   this.msgSequenceID    = 0;
   this.pause            = false;
   this.stop             = false;
   this.numberFormatter  = new DecimalFormat(NBR_FORMAT);

   // Load properties

   properties          = loadProperties(connectionProperties);
   addJndiICProperties = new HashMap<String, String>();

   // Set properties

   this.jndiFactoryInitial         = properties.getProperty("jndi.ic.factoryinitial",       null);
   this.jndiProviderURL            = properties.getProperty("jndi.ic.providerurl",          null);
   this.jndiConnectionFactoryAlias = properties.getProperty("jndi.alias.connectionfactory", null);
   this.jndiQueueAlias             = properties.getProperty("jndi.alias.queue",             null);

   for(String key : properties.stringPropertyNames())
   {
    if(key.startsWith("jndi.ic.add.")) addJndiICProperties.put(key.substring("jndi.ic.add.".length()), properties.getProperty(key));
   }

   this.addJndiICProperties = Collections.unmodifiableMap(addJndiICProperties);

   // Log parameters

   LOGGER.info("=== " + this.getClass().getSimpleName() + " :: " + name + " ===");
   LOGGER.info("JNDIFACTORYINITIAL         = " + jndiFactoryInitial);
   LOGGER.info("JNDIPROVIDERURL            = " + jndiProviderURL);
   LOGGER.info("JNDICONNECTIONFACTORYALIAS = " + jndiConnectionFactoryAlias);
   LOGGER.info("JNDIQUEUEALIAS             = " + jndiQueueAlias);
   LOGGER.info("ADDJNDIICPROPERTIES        = " + addJndiICProperties);
  }

  // === Public ================================================================

  /**
   * Connect the connector
   * @throws JMSException
   * @throws NamingException
   */

  abstract void connect()
  throws JMSException,
         NamingException;

  /**
   * Connect the connector with retries
   * @throws InterruptedException
   */

  void connectWithRetries()
  throws InterruptedException
  {
   // Loop until successful connect

   for(;;)
   {
    try
    {
     connect(); return;
    }
    catch(Exception e)
    {
     LOGGER.error(name + ": Exception while connecting; retrying.", e);

     disconnect();
     Thread.sleep(1000);
    }
   }
  }

  /**
   * Disconnect the connector
   */

  abstract void disconnect();

  /**
   * Reconnect the connector with retries
   * @throws InterruptedException
   */

  void reconnectWithRetries()
  throws InterruptedException
  {
   disconnect();
   connectWithRetries();
  }

  /**
   * @return
   * @throws NamingException
   */

  InitialContext createInitialContext()
  throws NamingException
  {
   Properties  jndiProperties;

   // Initialize JNDI

   jndiProperties = new Properties();

   // Add standard properties

   jndiProperties.setProperty(Context.INITIAL_CONTEXT_FACTORY, jndiFactoryInitial);
   jndiProperties.setProperty(Context.PROVIDER_URL,            jndiProviderURL);

   // Add additional properties

   for(Map.Entry<String, String> entry : addJndiICProperties.entrySet())
   {
    jndiProperties.setProperty(entry.getKey(), entry.getValue());
   }

   // Create and return context

   return(new InitialContext(jndiProperties));
  }

  /**
   * Change the connector pause status
   */

  void signalPause(boolean pause)
  {
   this.pause = pause;
  }

  /**
   * Stop the connector
   */

  void signalStop()
  {
   stop = true;
  }

  /**
   * @return the current message sequence ID
   */

  long getCurrentMessageSequenceID()
  {
   return(msgSequenceID);
  }

  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================
 }

 /**
  * Sender class
  */

 private class Sender extends Connector
 {
  // === Static ================================================================
  // === Attributes ============================================================

  private MessageProducer producer;

  // === Constructor ===========================================================

  /**
   * Create a new Sender instance
   * @param name                       the connector name
   * @param jndiFactoryInitial         the JNDI factory alias
   * @param jndiProviderURL            the JNDI provider URL
   * @param jndiConnectionFactoryAlias the JNDI connection factory alias
   * @param jndiQueueAlias             the JNDI queue alias
   * @param addJndiICProperties        any additional JNDI InitialContext properties
   */

  public Sender(final String name,
                final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                final Map<String, String> addJndiICProperties)
  {
   super(name, jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias, addJndiICProperties);
  }

  /**
   * Create a new Sender instance
   * @see Connector#Connector(String, File)
   */

  public Sender(final String name, final File connectionProperties)
  {
   super(name, connectionProperties);
  }

  // === Public ================================================================

  /**
   * Send messages until stopped.
   * @see java.lang.Runnable#run()
   */

  public void run()
  {
   boolean    retryBatch;
   int        retryID, batchSize;
   long       preBatchMsgSequenceID, lastMsgLogTime, firstBatchMsgSequenceID, lastBatchMsgSequenceID,
              batchStartTime;

   byte[]     payload;

   MapMessage mapMessage;

   try
   {
    // Initial connect

    connectWithRetries();

    // Try/Catch for InterruptedExceptions

    try
    {
     // Loop until stopped

     while(!stop)
     {
      // Save message sequence ID

      preBatchMsgSequenceID = msgSequenceID;

      // Loop until a batch has been sent and committed. Batch is being retried
      // in case of message creation, sending or commit failures. Duplicates can
      // be generated when retrying after commit failures.

      for(retryID = 0; ; retryID ++)
      {
       // Initialize

       retryBatch = false;

       // Set initial message sequence ID

       msgSequenceID = preBatchMsgSequenceID;

       // Log

       LOGGER.info(name + ": Sending ...");
       lastMsgLogTime = System.currentTimeMillis();

       // Create & send a batch of message

       try
       {
        // Initialize
        
        batchSize               =  0;
        firstBatchMsgSequenceID = -1L;
        lastBatchMsgSequenceID  = -1L;
        batchStartTime          = System.currentTimeMillis();
        
        // Send messages until we either reach the batch end time or the number of messages per
        // batch, depending on the batch mode
        
        while(SENDERBATCHMODE == BatchMode.TIME ? System.currentTimeMillis() - batchStartTime < SENDERCOMMITINTERVAL :
                                                  batchSize                                   < SENDERCOMMITINTERVAL)
        {
         // Update first/last batch message sequence ID and counter
         
         if(firstBatchMsgSequenceID < 0L) 
         {
          firstBatchMsgSequenceID = msgSequenceID;
         }
         
         lastBatchMsgSequenceID = msgSequenceID;
         
         batchSize ++;
         
         // Select payload

         payload = MESSAGES.get(RANDOM.nextInt(MESSAGES.size()));

         // Create and send message

         mapMessage = session.createMapMessage();

         mapMessage.setStringProperty(ID_MESSAGEID,  this.getClass().getSimpleName() + "-" + STARTTIME + "-" + name + "-" + msgSequenceID);
         mapMessage.setLongProperty  (ID_TIMESTAMP,  System.currentTimeMillis());
         mapMessage.setLongProperty  (ID_SEQUENCEID, msgSequenceID);
         mapMessage.setIntProperty   (ID_RETRYID,    retryID);
         mapMessage.setBytes         (ID_PAYLOAD,    payload);

         producer.send(mapMessage);

         // Log?

         if(lastMsgLogTime < System.currentTimeMillis() - MSG_LOG_INTERVAL)
         {
          LOGGER.info(name + ": Message # " + numberFormatter.format(msgSequenceID) + " sent.");
          lastMsgLogTime = System.currentTimeMillis();
         }

         // Increase message sequence ID

         msgSequenceID ++;
        }

        // Log

        LOGGER.info(name + ": Messages sent; batch: first: " + numberFormatter.format(firstBatchMsgSequenceID) + " last: " + numberFormatter.format(lastBatchMsgSequenceID) + " count: " + numberFormatter.format(batchSize) + "; totals: #msgs sent: " + numberFormatter.format(msgSequenceID));
        lastMsgLogTime = System.currentTimeMillis();
       }
       catch(JMSException e)
       {
        // Message creation or sending failed.
        // Log

        LOGGER.error(name + ": Message creation or sending failed, reconnecting and retrying.", e);

        // Retry batch

        retryBatch = true;
       }

       // Commit batch

       if(!retryBatch)
       {
        // Log

        LOGGER.info(name + ": Committing messages ...");

        // Try to commit

        try
        {
         // Commit

         session.commit();

         // Log

         LOGGER.info(name + ": Messages committed.");
        }
        catch(JMSException e)
        {
         // Log

         LOGGER.error(name + ": Commit failed, reconnecting and retrying.", e);

         // Retry batch

         retryBatch = true;
        }
       }

       // Retry? Reconnect first. Else break loop

       if(retryBatch) reconnectWithRetries();
                 else break;
      }

      // Control threads
      
      controlThreads();
      
      // Pause?

      while(pause && !stop) Thread.sleep(100);
     }
    }
    catch(InterruptedException e)
    {
     // Log

     LOGGER.error(name + ": Caught exception", e);
    }
   }
   catch(InterruptedException e)
   {
    // Log

    LOGGER.error(name + ": Interrupted", e);
   }
   finally
   {
    disconnect();
   }
  }

  /**
   * @see com.tomtomwork.webfleet.components.jms.test.MultiThreadedLoadTest.Connector#connect()
   */

  @Override
  public void connect()
  throws JMSException,
         NamingException
  {
   InitialContext    jndiContext;
   ConnectionFactory connectionFactory;
   Destination       destination;

   // Log

   LOGGER.info(name + ": Connecting.");

   // Initialize context

   jndiContext = createInitialContext();

   // Lookup connection factory and queue

   connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
   destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

   // Create producer connection, session and message producer

   connection = connectionFactory.createConnection();
   session    = connection.createSession(true, 0);

   producer = session.createProducer(destination);
   producer.setDeliveryMode(PERSISTANT_STORAGE ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT);

   connection.start();

   // Log

   LOGGER.info(name + ": Connected.");
  }

  /**
   * @see com.tomtomwork.webfleet.components.jms.test.MultiThreadedLoadTest.Connector#disconnect()
   */

  @Override
  public void disconnect()
  {
   // Log

   LOGGER.info(name + ": Disconnecting.");

   // Close assets

   if(producer   != null) { try { producer  .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting producer",   e); } producer   = null; }
   if(session    != null) { try { session   .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting session",    e); } session    = null; }
   if(connection != null) { try { connection.close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting connection", e); } connection = null; }

   // Log

   LOGGER.info(name + ": Disconnected.");
  }

  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================
 }

 /**
  * Receiver class
  */

 private class Receiver extends Connector
 {
  // === Static ================================================================
  // === Attributes ============================================================

  private final DynBitSet       receivedMessages, missingMessages;
  private       MessageConsumer consumer;

  // === Constructor ===========================================================

  /**
   * Create a new Receiver instance
   * @param name                       the connector name
   * @param jndiFactoryInitial         the JNDI factory alias
   * @param jndiProviderURL            the JNDI provider URL
   * @param jndiConnectionFactoryAlias the JNDI connection factory alias
   * @param jndiQueueAlias             the JNDI queue alias
   * @param addJndiICProperties        any additional JNDI InitialContext properties
   */

  public Receiver(final String name,
                  final String jndiFactoryInitial, final String jndiProviderURL, final String jndiConnectionFactoryAlias, final String jndiQueueAlias,
                  final Map<String, String> addJndiICProperties)
  {
   // Call super constructor

   super(name, jndiFactoryInitial, jndiProviderURL, jndiConnectionFactoryAlias, jndiQueueAlias, addJndiICProperties);

   // Initialize attributes

   this.receivedMessages = new DynBitSet(128 * 1024);
   this.missingMessages  = new DynBitSet(128 * 1024);
  }

  /**
   * Create a new Receiver
   * @see Connector#Connector(String, File)
   */

  public Receiver(final String name, final File connectionProperties)
  {
   // Call super constructor

   super(name, connectionProperties);

   // Initialize attributes

   this.receivedMessages = new DynBitSet(128 * 1024);
   this.missingMessages  = new DynBitSet(128 * 1024);
  }

  // === Public ================================================================

  /**
   * Send messages until stopped.
   * @see java.lang.Runnable#run()
   */

  public void run()
  {
   boolean    retryBatch;
   int        retryID, batchSize;
   long       receivedMsgSequenceID, sequenceErrors, duplicateErrors, lastMsgLogTime,
              firstBatchMsgSequenceID, lastBatchMsgSequenceID, batchStartTime, 
              batchTimeout;
   
   MapMessage mapMessage;

   // Initialize

   sequenceErrors  = 0L;
   duplicateErrors = 0L;
   lastMsgLogTime  = 0L;

   try
   {
    // Initial connect

    connectWithRetries();

    // Try/Catch for InterruptedExceptions

    try
    {
     // Loop until stopped

     while(!stop)
     {
      // Loop until a batch has been sent and committed. Batch is being retried
      // in case of message creation, sending or commit failures. Duplicates can
      // be generated when retrying after commit failures.

      for(retryID = 0; ; retryID ++)
      {
       // Initialize

       retryBatch = false;

       // Log

       LOGGER.info(name + ": Receiving ...");
       lastMsgLogTime = System.currentTimeMillis();

       // Create & send a batch of message

       try
       {
        // Initialize
        
        batchSize               =  0;
        firstBatchMsgSequenceID = -1L;
        lastBatchMsgSequenceID  = -1L;
        batchStartTime          = System.currentTimeMillis();
        
        // Send messages until we either reach the batch end time or the number of messages per
        // batch, depending on the batch mode
        
        ReceiveLoop: for(;;)
        {
         // Process based on batch mode
         
         switch(RECEIVERBATCHMODE)
         {
          case TIME  : batchTimeout = RECEIVERCOMMITINTERVAL - (System.currentTimeMillis() - batchStartTime);
                       if(batchTimeout <= 0) break ReceiveLoop;
                       break;
          
          case COUNT : batchTimeout = RECEIVE_TIMEOUT;
                       if(batchSize >= RECEIVERCOMMITINTERVAL) break ReceiveLoop;
                       break;
                       
          default    : throw new IllegalArgumentException("" + RECEIVERBATCHMODE);
         }
      
         // Update first/last batch message sequence ID and counter
         
         if(firstBatchMsgSequenceID < 0L) 
         {
          firstBatchMsgSequenceID = msgSequenceID;
         }
         
         lastBatchMsgSequenceID = msgSequenceID;
         
         batchSize ++;

         // Receive
         
         if((mapMessage = ((MapMessage)(consumer.receive(batchTimeout)))) != null)
         {
          // Get message sequence ID of received message

          receivedMsgSequenceID = mapMessage.getLongProperty(ID_SEQUENCEID);

          // Log?

          if(lastMsgLogTime < System.currentTimeMillis() - MSG_LOG_INTERVAL)
          {
           LOGGER.info(name + ": Message # " + numberFormatter.format(receivedMsgSequenceID) + " received.");
           lastMsgLogTime = System.currentTimeMillis();
          }

          // Check, if we already received this message

          if(receivedMessages.get(receivedMsgSequenceID))
          {
           // Duplicate message

           LOGGER.warn(name + ": Message # " + numberFormatter.format(receivedMsgSequenceID) + ": Duplicate.");
           lastMsgLogTime = System.currentTimeMillis();

           // Count

           duplicateErrors ++;
          }
          else
          {
           // Message not seen before. Mark message as being received
           // and not being missed (if missed before) anymore

           receivedMessages.set  (receivedMsgSequenceID);
           missingMessages .clear(receivedMsgSequenceID);
          }

          // Compare message sequence ID

          if(receivedMsgSequenceID != msgSequenceID)
          {
           // Received a different message than expected
           // Log error

           LOGGER.warn(name + ": Received message # " + numberFormatter.format(receivedMsgSequenceID) + ", expected " +  numberFormatter.format(msgSequenceID) + ".");
           lastMsgLogTime = System.currentTimeMillis();

           // Count sequence error

           sequenceErrors ++;

           // If we received a newer message than expected, check, if we already received
           // the messages in-between both IDs. If not, record these messages as potentially
           // missing.

           if(receivedMsgSequenceID > msgSequenceID)
           {
            for(long chkID = msgSequenceID; chkID < receivedMsgSequenceID; chkID ++)
            {
             if(!receivedMessages.get(chkID)) missingMessages.set(chkID);
            }
           }

           // Re-sync

           msgSequenceID = receivedMsgSequenceID;
          }

          // Increase message sequence ID

          msgSequenceID ++;
         }
         else
         {
          // No message received within timeout. Throw exception to reconnect and retry
          // if running in count based mode; break receive loop if running in time based
          // mode
          
          switch(RECEIVERBATCHMODE)
          {
           case TIME  : break ReceiveLoop;
           case COUNT : throw new JMSException(name + ": Timed out receiving messages.");
           default    : throw new IllegalArgumentException("" + RECEIVERBATCHMODE);
          }
         }
        }

        // Log

        LOGGER.info(name + ": Messages received; batch: first: " + numberFormatter.format(firstBatchMsgSequenceID) + " last: " + numberFormatter.format(lastBatchMsgSequenceID) + " count: " + numberFormatter.format(batchSize));
        LOGGER.info(name + ": Messages received; totals: #msgs received: " + numberFormatter.format(msgSequenceID) + ", #seq errors: " + sequenceErrors + "; #msg missing: " + missingMessages.getNbrSetBits() + "; #msg dupl: " + duplicateErrors);
        lastMsgLogTime = System.currentTimeMillis();
       }
       catch(JMSException e)
       {
        // Message reception failed.
        // Log

        LOGGER.error(name + ": Message reception failed, reconnecting and retrying.", e);

        // Retry batch

        retryBatch = true;
       }

       // Commit batch

       if(!retryBatch)
       {
        // Log

        LOGGER.info(name + ": Committing messages ...");

        // Try to commit

        try
        {
         // Commit

         session.commit();

         // Log

         LOGGER.info(name + ": Messages committed.");
        }
        catch(JMSException e)
        {
         // Log

         LOGGER.error(name + ": Commit failed, reconnecting and retrying.", e);

         // Retry batch

         retryBatch = true;
        }
       }

       // Retry? Reconnect first. Else break loop

       if(retryBatch) reconnectWithRetries();
                 else break;
      }
      
      // Control threads
      
      controlThreads();      

      // Pause?

      while(pause && !stop) Thread.sleep(100);
     }
    }
    catch(InterruptedException e)
    {
     // Log

     LOGGER.error(name + ": Caught exception", e);
    }
   }
   catch(InterruptedException e)
   {
    // Log

    LOGGER.error(name + ": Interrupted", e);
   }
   finally
   {
    disconnect();
   }
  }

  /**
   * @see com.tomtomwork.webfleet.components.jms.test.MultiThreadedLoadTest.Connector#connect()
   */

  @Override
  public void connect()
  throws JMSException,
         NamingException
  {
   InitialContext    jndiContext;
   ConnectionFactory connectionFactory;
   Destination       destination;

   // Log

   LOGGER.info(name + ": Connecting.");

   // Initialize context

   jndiContext = createInitialContext();

   // Lookup connection factory and queue

   connectionFactory = ((ConnectionFactory)(jndiContext.lookup(jndiConnectionFactoryAlias)));
   destination       = ((Destination      )(jndiContext.lookup(jndiQueueAlias)));

   // Create producer connection, session and message consumer

   connection = connectionFactory.createConnection();
   session    = connection.createSession(true, 0);

   consumer = session.createConsumer(destination);

   connection.start();

   // Log

   LOGGER.info(name + ": Connected.");
  }

  /**
   * @see com.tomtomwork.webfleet.components.jms.test.MultiThreadedLoadTest.Connector#disconnect()
   */

  @Override
  public void disconnect()
  {
   // Log

   LOGGER.info(name + ": Disconnecting.");

   // Close assets

   if(consumer   != null) { try { consumer  .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting consumer",   e); } consumer   = null; }
   if(session    != null) { try { session   .close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting session",    e); } session    = null; }
   if(connection != null) { try { connection.close(); } catch(JMSException e) { LOGGER.error("Error while disconnecting connection", e); } connection = null; }

   // Log

   LOGGER.info(name + ": Disconnected.");
  }

  // === Package protected =====================================================
  // === Protected =============================================================
  // === Private ===============================================================
  // === Inner classes =========================================================
 }
}
